/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.  
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/
/*
* Modifications: 
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

 * Modification author: Hejun Wu (vmnet@cs.ust.hk)
 
********************************************************************************/
/*
sensor hardware emulation module
*/
#include <math.h>
#include "adc.h"
#include "../EM/atmega128.h"
/*
* This module implements the basic behavior of the Analog to Digital
* Converter.  IO Registers required:
*
* MUX - ADC Multiplexer Selection Register
*    Bind to ADMUX, mask 0xFF
* CONTROL - ADC Control and Status Register A
*    Bind to ADCSRA, mask 0xFF
* DATALOW - ADC Data Register Low
*    Bind to ADCL, mask 0xFF
* DATAHIGH - ADC Data Register High
*    Bind to ADCH, mask 0xFF
*
* Interrupts are defined:
*
* SAMPLE - Triggered when a sample is completed
*    Bind to 22 ($002A)
*
* Parameters:
*
* AVCC - Voltage at the Analog power source (defaults to 3.3V)
* AREF - Voltage at the Analog reference pin (defaults to AVCC)
*/

/* Declare local static functions */
static void adc_destroy_device(void *device);
static error_t adc_enter_sleep(void *device, uint8_t sleepmode);
static error_t adc_leave_sleep(void *device);
static int adc_clock_tick(void *device);
static int adc_clearint_sample(void *device);
uint8_t adc_control_read(void *dev);
void adc_control_write(void *dev, uint8_t value);
uint8_t adc_mux_read(void *dev);
void adc_mux_write(void *dev, uint8_t value);
uint8_t adc_datalow_read(void *dev);
void adc_datalow_write(void *dev, uint8_t value);
uint8_t adc_datahigh_read(void *dev);
void adc_datahigh_write(void *dev, uint8_t value);

//////////////////////////////////////////////////////////////////////////////////////////////////////
//VMNet:
//The following is for power consumption logging
#define ADC_OPERATION_TYPES    4

unsigned char strADCOperations[ADC_OPERATION_TYPES][MAX_OPERATION_STR_LEN]={"ADC:SLEEP","ADC:CONTROL","ADC:SAMPLE","ADC:IDLE"};

//////////////////////////////////////////////////////////////////////////////////////////////////////

/*
* Startup functions
*/

device_obj* create_device(char *name) {
    unsigned i;
    adc_device *new_dev = (adc_device*)malloc(sizeof(adc_device));
    if(new_dev == NULL) {
        return NULL;
    }
	
    memset(new_dev, 0, sizeof(adc_device));
	
    new_dev->device.destroy = adc_destroy_device;
    new_dev->device.sleep = adc_enter_sleep;
    new_dev->device.wake = adc_leave_sleep;
    new_dev->device.dev = new_dev;
    new_dev->clk_tick.tick = NULL;
    new_dev->clk_tick.dev = new_dev;
    new_dev->int_sample.clear = adc_clearint_sample;
    new_dev->int_sample.dev = new_dev;
	
    new_dev->reg_mux.read = adc_mux_read;
    new_dev->reg_mux.write = adc_mux_write;
    new_dev->reg_mux.dev = new_dev;
    new_dev->reg_control.read = adc_control_read;
    new_dev->reg_control.write = adc_control_write;
    new_dev->reg_control.dev = new_dev;
    new_dev->reg_datalow.read = adc_datalow_read;
    new_dev->reg_datalow.write = adc_datalow_write;
    new_dev->reg_datalow.dev = new_dev;
    new_dev->reg_datahigh.read = adc_datahigh_read;
    new_dev->reg_datahigh.write = adc_datahigh_write;
    new_dev->reg_datahigh.dev = new_dev;
	
    /* Default power supply to 3.3 Volts */
    new_dev->avcc = 3.3;
    /* Default external analog reference to AVCC */
    new_dev->aref = 3.3;
    new_dev->aref_set = 0;
    /* Data registers not locked */
    new_dev->datalocked = 0;
	
    /* No sources */
    for(i = 0; i < 8; i++) {
        new_dev->sources[i] = NULL;
    }
	
    return &(new_dev->device);
}

register_obj *add_register(void *dev, char *name) {
    adc_device *adc = (adc_device*)dev;
	
    if(strcasecmp(name, "MUX") == 0) {
        return &(adc->reg_mux);
    } else if(strcasecmp(name, "CONTROL") == 0) {
        return &(adc->reg_control);
    } else if(strcasecmp(name, "DATALOW") == 0) {
        return &(adc->reg_datalow);
    } else if(strcasecmp(name, "DATAHIGH") == 0) {
        return &(adc->reg_datahigh);
    }
	
    return NULL;
}

interrupt_obj* add_interrupt(void *dev, char *name) {
    adc_device *adc = (adc_device*)dev;
	
    if(strcasecmp(name, "SAMPLE") == 0) {
        return &(adc->int_sample);
    }
	
    return NULL;
}

clock_obj* add_clock(void *dev) {
    adc_device *adc = (adc_device*)dev;
    return &(adc->clk_tick);
}

error_t set_param(void *dev, char *key, char *value) {
    adc_device *adc = (adc_device*)dev;
    double parse;
    char *endparse;
	
    if(strcasecmp(key, "AVCC") == 0) {
        parse = strtod(value, &endparse);
        if(endparse && *endparse) {
            /* Non-numeric character */
            adc->device.device_debug(&adc->device, V_DEBUG, "Invalid floating point number provided for AVCC.\n");
            return -1;
        } else if(parse <= 0) {
            /* Illegal AVCC */
            adc->device.device_debug(&adc->device, V_DEBUG, "AVCC must be greater than 0 Volts.\n");
            return -1;
        }
        adc->avcc = parse;
        if(!adc->aref_set) {
            adc->aref = parse;
        }
        return 0;
    } else if(strcasecmp(key, "AREF") == 0) {
        parse = strtod(value, &endparse);
        if(endparse && *endparse) {
            /* Non-numeric character */
            adc->device.device_debug(&adc->device, V_DEBUG, "Invalid floating point number provided for AREF.\n");
            return -1;
        } else if(parse <= 0) {
            /* Illegal AVCC */
            adc->device.device_debug(&adc->device, V_DEBUG, "AREF must be greater than 0 Volts.\n");
            return -1;
        }
        adc->aref = parse;
        adc->aref_set = 1;
        return 0;
    }
	
    return -1;
}

/*
* Basic device callbacks
*/

void adc_destroy_device(void *dev) {
    adc_device *adc = (adc_device*)dev;
    free(adc);
}

error_t adc_enter_sleep(void *dev, uint8_t sleepmode) {
    adc_device *adc = (adc_device*)dev;
	
    switch(sleepmode) {
    case SLEEPMODE_IDLE:
    case SLEEPMODE_ADCNOISE:
        /* Sleep modes that start a conversion */
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        LOG_CURRENT_OPERATION(ADC_IDLE,ADC_MASK);
		//+++++++++++++++++++++++++++++++++++++++++++++++
        if(!GET_FLAG(adc->control, ADC_CONTROL_ADSC)
			&& GET_FLAG(adc->control, ADC_CONTROL_ADEN)) {
            adc->clockcount = 0;
            device_setclock(adc->device, adc->clk_tick, adc_clock_tick);
        }
        break;
    default:
        /* Everything else shuts us down */          
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        LOG_CURRENT_OPERATION(ADC_SLEEP,ADC_MASK);   
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        device_clrclock(adc->device, adc->clk_tick);
        adc->clockcount = 0;
        break;
    }
    return 0;
}

error_t adc_leave_sleep(void *dev) {
    adc_device *adc = (adc_device*)dev;
    /* If we are enabled, reenter conversion */
    if(GET_FLAG(adc->control, ADC_CONTROL_ADSC)
		&& GET_FLAG(adc->control, ADC_CONTROL_ADEN)) {
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK); 
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        device_setclock(adc->device, adc->clk_tick, adc_clock_tick);
    }
    return 0;
}

int adc_clearint_sample(void *dev) {
    adc_device *adc = (adc_device*)dev;
    CLR_FLAG(adc->control, ADC_CONTROL_ADIF);        
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
    return 0;
}

/*
* Device I/O
*/

uint8_t adc_control_read(void *dev) {
    adc_device *adc = (adc_device*)dev;              
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
    return adc->control;
}

void adc_control_write(void *dev, uint8_t value) {
    adc_device *adc = (adc_device*)dev;
    static const uint8_t prescalers[] = { 2, 2, 4, 8, 16, 32, 64, 128 };
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
    /* See if the ADC is being disabled */
    if(GET_FLAG(adc->control, ADC_CONTROL_ADEN)) {
        if(!GET_FLAG(value, ADC_CONTROL_ADEN)) {
            /* Being disabled */
            device_clrclock(adc->device, adc->clk_tick);
            adc->clockcount = 0;
            adc->device.device_debug(&adc->device, V_DEBUG, "ADC unit disabled\n");
        }
    } else {
        if(GET_FLAG(value, ADC_CONTROL_ADEN)) {
            /* Being enabled */
            device_clrclock(adc->device, adc->clk_tick);
            adc->clockcount = 0;
            /* First conversion takes longer */
            adc->clocksample = 14;
            adc->clockcomplete = 26;
            adc->device.device_debug(&adc->device, V_DEBUG, "ADC unit enabled\n");
        }
    }
    /* See if a conversion is starting */
    if(GET_FLAG(value, ADC_CONTROL_ADSC)
		&& GET_FLAG(value, ADC_CONTROL_ADEN)) {
        device_setclock(adc->device, adc->clk_tick, adc_clock_tick);
        adc->device.device_debug(&adc->device, V_DEBUG, "Beginning ADC conversion\n");
    } else {
        /* Writing to ADSC has no effect during a conversion */
        value = (value & ~ADC_CONTROL_ADSC)
            | (adc->control & ADC_CONTROL_ADSC);
    }
    /* See if we clear an interrupt */
    if(GET_FLAG(value, ADC_CONTROL_ADIF)) {
        CLR_FLAG(value, ADC_CONTROL_ADIF);
    } else {
        /* Copy over the old interrupt flag */
        if(GET_FLAG(adc->control, ADC_CONTROL_ADIF)) {
            SET_FLAG(value, ADC_CONTROL_ADIF);
        }
    }
    /* See if we signal an interrupt */
    if(GET_FLAG(value, ADC_CONTROL_ADIE)
		&& GET_FLAG(value, ADC_CONTROL_ADIF)) {
        adc->int_sample.set_interrupt(& adc->int_sample);
    }
	
    /* Set prescaler and copy new control value */
    adc->prescaler = prescalers[value & ADC_CONTROL_ADPS];
    adc->control = value;
}

uint8_t adc_mux_read(void *dev) {
    adc_device *adc = (adc_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
    return adc->mux;
}

void adc_mux_write(void *dev, uint8_t value) {
    adc_device *adc = (adc_device*)dev;
    adc->mux = value;                                
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
}

uint8_t adc_datalow_read(void *dev) {
	
    adc_device *adc = (adc_device*)dev;
    adc->datalocked = 1;                             
	//VMNet data source++++++++++++++++++++++++++++++++++
	adc_read_sensory_data(adc);
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
	//+++++++++++++++++++++++++++++++++++++++++++++++++++
    return adc->datalow;
}

void adc_datalow_write(void *dev, uint8_t value) {
    adc_device *adc = (adc_device*)dev;
    adc->device.device_debug(&adc->device, V_DEBUG, "Crazy user is writing to a read-only register.\n");
}

uint8_t adc_datahigh_read(void *dev) {
    adc_device *adc = (adc_device*)dev;
    adc->datalocked = 0;                             
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
	
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
    return adc->datahigh;
}

void adc_datahigh_write(void *dev, uint8_t value) {
    adc_device *adc = (adc_device*)dev;
    adc->device.device_debug(&adc->device, V_DEBUG, "Crazy user is writing to a read-only register.\n");
}

/*
* Public interface
*/

double adc_read_multiplexer(adc_device* device) {
    /* Positive end selector */
    static const int pos_in [] = {
        /* Single Ended Input */
        0, 1, 2, 3, 4, 5, 6, 7,
			/* Differential */
			0, 1, 0, 1, 2, 3, 2, 3,
			0, 1, 2, 3, 4, 5, 6, 7,
			0, 1, 2, 3, 4, 5,
			/* 1.23V Internal Band Gap */
			8,
			/* 0V Analog Ground */
			9
    };
    int muxread = pos_in[device->mux & ADC_MUX_MUX];
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(ADC_CONTROL,ADC_MASK);
    if(muxread < 8) {
        if(device->sources[muxread] != NULL) {
            return device->sources[muxread](device->sources_dev[muxread],
				device->avcc);
        } else {
            device->device.device_debug(&device->device, V_DEBUG, "Read from unconnected input ADC%d\n", muxread);
            return 0.0;
        }
    } else if(muxread == 8) {
        return 1.23;
    } else {
        return 0.0;
    }
}


error_t adc_set_source(adc_device* device, int sourcenum,
					   adc_sample_callback func, void *pvoid) {
    if((sourcenum < 0) || (sourcenum > 7)) {
        device->device.device_event(&device->device, "Attempt to assign to illegal device %d!\n", sourcenum);
        return -1;
    }
	
    if(device->sources[sourcenum] != NULL) {
        return -1;
    }
	
    device->sources[sourcenum] = func;
    device->sources_dev[sourcenum] = pvoid;
    return 0;
}

//////////////////////////////////////////////////////////////////////////////////
//ADC read source data  adc_read_sensory_data
//////////////////////////////////////////////////////////////////////////////////

#ifdef SENSORY_DATA_LOG
int log_sensory_epochs()
{
#if 0
    if((g_sample_data_count[g_idChip]/g_num_attributes)==MAX_SENSOR_DATA_SIZE)
    {
        dump_sensory_data_log();
        memset(g_sensory_data_rec,0,SENSOR_DATA_REC_LEN);
	}
    sprintf(g_sensory_data_rec[g_idChip],"%d\n",g_chips[g_idChip]->cycles);
#else
    
#endif 
    
}
#else
#define log_sensory_epochs()
#endif
#define ADC_ENABLED_ATTRIBUTES 8
#define REAL_ADC_DATA
int adc_read_sensory_data(void *device)
{
	
    adc_device *adc = (adc_device*)device;
    int i,epochs;
	
    unsigned int sensory_data_count=0;
#if 1
if(g_sensor_data)
	if(g_sensor_data[g_idChip])
	{
		unsigned short tempData;
		if(g_startCycles)
		g_sample_data_count[g_idChip]=(g_chips[0]->cycles-g_startCycles)/73728*10/g_sample_period;
		else
			g_sample_data_count[g_idChip]=0;
		sensory_data_count=(unsigned int)(g_sample_data_count[g_idChip]*MAX_SENSOR_ATTRIBUTES) % (MAX_SENSOR_DATA_SIZE*MAX_SENSOR_ATTRIBUTES);
		if(sensory_data_count>(MAX_SENSOR_DATA_SIZE*MAX_SENSOR_ATTRIBUTES))
		{
            printf("ERROR: not that much data\n");
            exit(-1);
			
		}
		if(g_sensor_data[g_idChip])
			tempData = g_sensor_data[g_idChip][sensory_data_count];
		else
			tempData = 518;
		//printf("SENSOR DATA=%d \n",tempData );
		adc->datalow = tempData&0xff;
		adc->datahigh = ((tempData&0xff00)>>8)&0xff;
	}

	else
#endif
	{
		adc->datalow =0x06;
		adc->datahigh =0x2;
	}
	
	
    
}

/*
* The ADC magic
*/
static int adc_clock_tick(void *device) {
    adc_device *adc = (adc_device*)device;
    unsigned bit;
    /* Negative input for differential mode */
    static const int neg_in[] = {
        /* Single Ended Input */
        -1, -1, -1, -1, -1, -1, -1, -1,
			/* Differential Mode */
			0,  0,  0,  0,  2,  2,  2,  2,
			1,  1,  1,  1,  1,  1,  1,  1,
			2,  2,  2,  2,  2,  2,
			/* Special inputs */
			-1, -1
    };
    /* Gains for differential mode */
    static const int gains[] = {
        /* Single Ended Input */
        1,   1,   1,   1,   1,   1,   1,   1,
			/* Differential Mode */
			
			10,  10,  200, 200, 10,  10,  200, 200,
			1,   1,   1,   1,   1,   1,   1,   1,
			1,   1,   1,   1,   1,   1,
			/* Special inputs */
			1,   1
    };
    /* Single Ended or Positive Input */
    static const int pos_in [] = {
        /* Single Ended Input */
        0, 1, 2, 3, 4, 5, 6, 7,
			/* Differential */
			0, 1, 0, 1, 2, 3, 2, 3,
			0, 1, 2, 3, 4, 5, 6, 7,
			0, 1, 2, 3, 4, 5,
			/* 1.23V Internal Band Gap */
			8,
			/* 0V Analog Ground */
			9
    };
	
	
    /* Clock generator */
    adc->clockin++;
    bit = GET_FLAG(adc->clockin, adc->prescaler);
    if(bit == adc->clockout) {
        /* Nothing happened... */
        return 0;
    } else {
        adc->clockout = bit;
    }

    /* Increment the cycle count on rising edge */
    if(bit) {
        adc->clockcount++;
    }
    /* Sample on falling edge of cycle #2 (or #14 on first conversion) */
    if(!bit && (adc->clockcount == adc->clocksample)) {
		//VMNet: for optimizaiton ++++++++++++++++++++++++++++++++++++++++++
#ifndef OPERATION_OPTIMIZE
        double samplepos, sampleneg, sampleref, sample;
        int posmux = pos_in[adc->mux & ADC_MUX_MUX];
        int negmux = neg_in[adc->mux & ADC_MUX_MUX];
        int gain = gains[adc->mux & ADC_MUX_MUX];
		
        /* Read the inputs */
        samplepos = adc_read_multiplexer(adc);
        if(negmux != -1) {
            if(adc->sources[negmux] != NULL) {
                sampleneg = adc->sources[negmux](adc->sources_dev[negmux],
					adc->avcc);
				
            } else {
                sampleneg = 0;
            }
        } else {
            sampleneg = 0;
        }
		
        /* Bound the inputs */
        adc->device.device_debug(&adc->device, V_DEBUG, "Read a voltage of %.3fV from ADC%d\n", samplepos, posmux);
        if(samplepos < 0) {
            adc->device.device_debug(&adc->device, V_DEBUG, "Input ADC%d under range!!!\n", posmux);
            samplepos = 0;
        } else if(samplepos > adc->avcc) {
            adc->device.device_debug(&adc->device, V_DEBUG, "Input ADC%d over range!!!\n", posmux);
            samplepos = adc->avcc;
        }
		
        if(negmux != -1) {
            adc->device.device_debug(&adc->device, V_DEBUG, "Read a voltage of %.3fV from ADC%d\n", sampleneg, negmux);
        }
        if(sampleneg < 0) {
            adc->device.device_debug(&adc->device, V_DEBUG, "Input ADC%d under range!!!\n", negmux);
            sampleneg = 0;
        } else if(sampleneg > adc->avcc) {
            adc->device.device_debug(&adc->device, V_DEBUG, "Input ADC%d over range!!!\n", negmux);
            sampleneg = adc->avcc;
        }
        /* Acquire our reference */
        switch(adc->mux & ADC_MUX_REFS) {
        case ADC_MUX_REFS_AREF:
            sampleref = adc->aref;
            break;
        case ADC_MUX_REFS_AVCC:
            sampleref = adc->avcc;
            break;
        case ADC_MUX_REFS_256V:
        default:
            sampleref = 2.56;
            break;
        }
		
        /* Perform the A->D */
		
        if(negmux == -1) {
            sample = floor(((samplepos * 1024) / sampleref) + 0.5);
            /* Check for saturation of the ADC */
            adc->sampleread = (sample < 1024) ? (uint16_t)sample : 1024;
        } else {
            sample = (samplepos - sampleneg) * gain;
            sample = floor(((sample * 512) / sampleref) + 0.5);
            /* Check for saturation in either direction */
            if(sample < -512) {
                sample = -512;
            } else if(sample > 511) {
                sample = 511;
            }
            /* Record sample */
            adc->sampleread = (uint16_t)sample;
        }
#else             
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        //loging the operation
        LOG_CURRENT_OPERATION(ADC_SAMPLE,ADC_MASK);
#endif
    }
	
    /* Report completion on rising edge of cycle #14 (or #26 on first) */
    if(bit && (adc->clockcount == adc->clockcomplete)) {
		
        adc->clockcount = 0;
        /* Make sure we are in normal operating mode now */
        adc->clocksample = 2;
        adc->clockcomplete = 14;                     
		//VMNet: optimizaiton ++++++++++++++++++++++++++++++++++++++++++
#ifndef OPERATION_OPTIMIZE
        adc->device.device_debug(&adc->device, V_DEBUG, "Reporting a reading of %d\n", adc->sampleread);
        /* If not locked, copy our value over */
        if(!adc->datalocked) {
            if(GET_FLAG(adc->mux, ADC_MUX_ADLAR)) {
                /* Left adjust */
                adc->datalow = ((adc->sampleread & 0x0003) << 6);
                adc->datahigh = ((adc->sampleread & 0x03FC) >> 2);
            } else {
                /* Right adjust */
                adc->datalow = (adc->sampleread & 0x00FF);
                adc->datahigh = ((adc->sampleread & 0x03FF) >> 8);
            }
            adc->device.device_debug(&adc->device, V_DEBUG, "Data High = 0x%02X, Data Low = 0x%02X\n",
				adc->datahigh, adc->datalow);
        }
#else
		//VMNet log++++++++++++++++++++++++++++++++++++++++++
        LOG_CURRENT_OPERATION(ADC_SAMPLE,ADC_MASK);
		//VMNet costant save time++++++++++++++++++++++++++++++++++++++++++
#endif
        /* Signal interrupt */
        SET_FLAG(adc->control, ADC_CONTROL_ADIF);
        if(GET_FLAG(adc->control, ADC_CONTROL_ADIE)) {
            adc->int_sample.set_interrupt(& adc->int_sample);
        }
        /* If we aren't in continuous mode, clear the start conversion */
        if(!GET_FLAG(adc->control, ADC_CONTROL_ADFR)) {
            CLR_FLAG(adc->control, ADC_CONTROL_ADSC);
        }
        /* If we aren't in free running mode, don't do another conversion */
        if(!GET_FLAG(adc->control, ADC_CONTROL_ADSC)) {
            device_clrclock(adc->device, adc->clk_tick);
        }
    }
    return 0;
}
