/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include "../EM/debug.h"

/* Structure for cpuex */
typedef struct patch_obj {
    /* Put next first since this might become variable length eventually */
    struct patch_obj *next;
    /* Address to apply patch to (byte address) */
    int32_t address;
    /* New value for location */
    uint8_t value;
} patch_obj;

typedef struct {
    device_obj device;
	
    /* List of memory addresses to patch */
    patch_obj *patchlist;
} cpuex_device;

/* Static functions */
static void cpuex_destroy_device(void *dev);
static error_t cpuex_wake(void *dev);
static error_t cpuex_patch(cpuex_device *dev, uint32_t addr, uint8_t value);

/*
* Startup functions
*/
device_obj* create_device(char *name) {
    cpuex_device *new_dev = (cpuex_device*)malloc(sizeof(cpuex_device));
    if(new_dev == NULL) {
        return NULL;
    }
	
    memset(new_dev, 0, sizeof(cpuex_device));
    new_dev->device.destroy = cpuex_destroy_device;
    new_dev->device.sleep = NULL;
    new_dev->device.wake = cpuex_wake;
    new_dev->device.dev = new_dev;
    new_dev->patchlist = NULL;
	
    return (device_obj*)new_dev;
}

/*
* Basic device callbacks
*/

static void cpuex_destroy_device(void *dev) {
    /* TODO: Free patch list */
    free(dev);
}

static error_t cpuex_wake(void *dev) {
    cpuex_device *cpuex = (cpuex_device*)dev;
	
    /* Execute patch list */
    if(cpuex->device.chip->flash != NULL) {
        patch_obj *patch = cpuex->patchlist, *next;
        cpuex->patchlist = NULL;
		
        while(patch) {
            cpuex_patch(cpuex, patch->address, patch->value);
            next = patch->next;
            free(patch);
            patch = next;
        }
    }
	
    return 0;
}

error_t set_param(void *dev, char *key, char *value) {
    cpuex_device *device = (cpuex_device*)dev;
	
    if(strcasecmp(key, "delay") == 0) {
        uint64_t parse;
        char *endparse;
		
        parse = strtoull(value, &endparse, 0);
        if(endparse && *endparse) {
            /* Non-numeric character */
            return -1;
        }
        /* Staggared start-up */
        device->device.chip->cycles = parse;
        return 0;
    } else if(strcasecmp(key, "patch") == 0) {
        char *tok, *endparse;
        int32_t address, data;
		
        /* Get address */
        tok = strtok(value, ",=");
        if((tok == NULL) || !*tok) {
            return 0;
        }
        address = strtoul(tok, &endparse, 0);
        if(endparse && *endparse) {
            /* Non-numeric character */
            fprintf(stderr, "Symbolic addresses not accepted for cpuex:patch (%s not numeric)\n", tok);
            return -1;
        }
        if(address >= (device->device.chip->flash_words * 2)) {
            /* Address too large */
            fprintf(stderr, "Address %s out of range, device only has %u bytes of flash\n", tok, device->device.chip->flash_words*2);
        }
		
        /* Get data */
        tok = strtok(NULL, ",=");
        if((tok == NULL) || !*tok) {
            return 0;
        }
        data = strtoul(tok, &endparse, 0);
        if(endparse && *endparse) {
            /* Non-numeric character */
            fprintf(stderr, "cpuex:patch only accepts numeric bytes at the moment (%s not numeric)\n", tok);
            return -1;
        }
        if(data > 255) {
            /* Too large */
            fprintf(stderr, "cpuex:patch only accepts numeric bytes at the moment (%s larger than 0xFF)\n", tok);
            return -1;
        }
		
        /* Store */
        cpuex_patch(device, address, (uint8_t)data);
        return 0;
    }
	
    return -1;
}

static error_t cpuex_patch(cpuex_device *dev, uint32_t addr, uint8_t value) {
    if(dev->device.chip->flash == NULL) {
        /* Queue */
        patch_obj *patch = (patch_obj*)malloc(sizeof(patch_obj));
        if(patch == NULL) {
            return -errno;
        }
        patch->address = addr;
        patch->value = value;
        patch->next = dev->patchlist;
        dev->patchlist = patch;
        return 0;
    } else {
        /* Patch */
        fprintf(stdout, "\tPatching memory location 0x%06x with value 0x%02x (was 0x%02x)\n",
			addr, value, ((uint8_t*)(dev->device.chip->flash))[addr]);
        ((uint8_t*)(dev->device.chip->flash))[addr] = value;
        return 0;
    }
}
