/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_SPI_H
#define ATEMU_SPI_H

#include "../EM/device.h"
#include "../EM/debug.h"

/* Bit positions in SPI Control Register */
#define SPI_SPCR_SPIE     0x80
#define SPI_SPCR_SPE      0x40
#define SPI_SPCR_DORD     0x20
#define SPI_SPCR_MSTR     0x10
#define SPI_SPCR_CPOL     0x08
#define SPI_SPCR_CPHA     0x04
#define SPI_SPCR_SPR1     0x02
#define SPI_SPCR_SPR0     0x01
#define SPI_SPCR_SPR      (SPI_SPCR_SPR1|SPI_SPCR_SPR0)
/* MSB first */
#define SPI_SPCR_DORD_MSB        0x00
/* LSB first */
#define SPI_SPCR_DORD_LSB        SPI_SPCR_DORD
/* Master select */
#define SPI_SPCR_MSTR_MASTER     SPI_SPCR_MSTR
/* Slave select */
#define SPI_SPCR_MSTR_SLAVE      0x00
/* Clock rises on leading edge */
#define SPI_SPCR_CPOL_LEADRISE   0x00
/* Clock falls on leading edge */
#define SPI_SPCR_CPOL_LEADFALL   SPI_SPCR_CPOL
/* Sample bit on leading edge */
#define SPI_SPCR_CPHA_LEADSAMPLE 0x00
/* Setup bit on leading edge */
#define SPI_SPCR_CPHA_LEADSETUP  SPI_SPCR_CPHA

/* Bit positions in SPI Status Register */
#define SPI_SPSR_SPIF     0x80
#define SPI_SPSR_WCOL     0x40
#define SPI_SPSR_RESERVED 0x3E
#define SPI_SPSR_SPI2X    0x01

/* Bit positions in Port B */
#define SPI_PORTB_SS      0x01 /* Slave Select */
#define SPI_PORTB_SCK     0x02 /* Serial Clock */
#define SPI_PORTB_MOSI    0x04 /* Master Out, Slave In */
#define SPI_PORTB_MISO    0x08 /* Master In, Slave Out */

typedef struct spi_device {
    device_obj    device;
	
    /* SPI Control Register */
    register_obj  reg_control;
    /* SPI Status Register */
    register_obj  reg_status;
    /* SPI Data Register */
    register_obj  reg_data;
	
    /* Port B interface */
    register_obj  reg_ddrb;
    register_obj  reg_portb;
    register_obj  reg_pinb;
	
    /* SPI Interrupt Request */
    interrupt_obj int_spi;
    /* Yes, we need a clock */
    clock_obj     clk_tick;
	
    /* Port B buffers */
    uint8_t portb_trueddr;
    uint8_t portb_ddr;
    uint8_t portb_out;
	
    /* SPI configuration */
    uint8_t spi_control;
	/* SPI shift register (right shift in, left shift out) */
    uint8_t spi_register;
	/* SPI input buffer */
    uint8_t spi_buffer;
	/* Number of shifts in the current transfer */
    uint8_t spi_shifts;
	/* TRUE if the SPI is enabled */
    unsigned spi_enabled:1;
	/* TRUE if interrupts are enabled */
    unsigned spi_intmask:1;
	/* TRUE if an interrupt has occurred */
    unsigned spi_intflag:1;
	/* TRUE if a write collision has occurred */
    unsigned spi_collision:1;
	/* Set to TRUE if oscillator is halved from SPCR_SPR */
    unsigned spi_spi2x:1;
	/* Collision is not cleared until both data and status are read */
    unsigned spi_colread:1;
	
    /* Pin interface */
	/* SCLK interface */
    pin_drive_enum pin_clock, pin_clockin;
	/* Master In, Slave Out interface */
    pin_drive_enum pin_miso, pin_misoin;
	/* Master Out, Slave In interface */
    pin_drive_enum pin_mosi, pin_mosiin;
	/* Slave Select interface */
    pin_drive_enum pin_slave, pin_slavein;
	
	/* Previous clock state */
    int pin_clock_prev;
	/* Previous Slave Select state */
    int pin_slave_prev;
	/* In some modes, we sense before we drive, so if we would shift,
	* we don't know what we need to drive. */
    int pin_lastread;
	
    /* Pointers to the functions to notify on pin driving */
    pin_drive_callback func_clock;
    void* func_clock_dev;
    pin_drive_callback func_miso;
    void* func_miso_dev;
    pin_drive_callback func_mosi;
    void* func_mosi_dev;
    pin_drive_callback func_slave;
    void* func_slave_dev;
} spi_device;

/* When the SPI has these pins set as inputs, they can be driven by
* external hardware.  Any device with a pointer to a SPI device (see the
* config option USES) can call this function to set the pin high or low.
* Attempting to drive a pin when it's set to output will result in
* kicking or screaming. */
error_t spi_drive_slave(spi_device* device, pin_drive_enum value);
error_t spi_drive_miso(spi_device* device, pin_drive_enum value);
error_t spi_drive_mosi(spi_device* device, pin_drive_enum value);
error_t spi_drive_clk(spi_device* device, pin_drive_enum value);
/* typedef, for type safety encouragement in dependant modules */
typedef error_t (*spi_drive_func)(spi_device* device, pin_drive_enum value);

/* If you want to be able to read from the structure without having to
* perform a callback, these functions may be useful.  Be careful, changes
* to the above structure may require recompiling of all dependant modules. */
static inline pin_drive_enum spi_read_slave(spi_device* device) {
    return device->pin_slave;
}
static inline pin_drive_enum spi_read_miso(spi_device* device) {
    return device->pin_miso;
}
static inline pin_drive_enum spi_read_mosi(spi_device* device) {
    return device->pin_mosi;
}
static inline pin_drive_enum spi_read_clock(spi_device* device) {
    return device->pin_clock;
}

/* If a device wants to receive change notifications whenever the value of
* one of the output pins changes, they can call this function to be notified
* on changes.  Only one device can be listening on a pin, so external
* hardware should share this callback with other devices.  If a second
* attempt to set a callback is made, -1 will be returned and it will not be
* set. */
error_t spi_set_slave(spi_device* device, pin_drive_callback func, void *pvoid);
error_t spi_set_miso(spi_device* device, pin_drive_callback func, void *pvoid);
error_t spi_set_mosi(spi_device* device, pin_drive_callback func, void *pvoid);
error_t spi_set_clk(spi_device* device, pin_drive_callback func, void *pvoid);
/* typedef for dependant modules */
typedef error_t (*spi_set_func)(spi_device* device, pin_drive_callback func, void *pvoid);

#endif /* ATEMU_SPI_H */
